/* vim: set tabstop=8 softtabstop=2 shiftwidth=2 expandtab: */
/*
 *    Copyright (C) 2011 Fabian Deutsch
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License along
 *    with this program; if not, write to the Free Software Foundation, Inc.,
 *    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

using Media;

public class MainClientWindow : GLib.Object
{
    Mx.Style mx_style;

    Gtk.Window window;
    GtkClutter.Embed embed;

    Clutter.Stage stage;
    Clutter.BinLayout layout;

    RTSPClient rtsp_client;
    Clutter.Texture video_texture;

    Mx.BoxLayout secondaries;

    ApplicationSettings application_settings;

    SettingsDialog settings_dialog;
    PeersDialog peers_dialog;
    VideoSettingsDialog video_settings_dialog;
    OneClickStreamingDialog oneclick_dialog;
    
    ISessionManager sessionmanager;
    uint sessionmanager_cookie;

    public string rtsp_address { get; set; }

    public bool decorated { get; set; }

    public MainClientWindow (string[] args, ApplicationSettings s)
    {
        GtkClutter.init (ref args);

        this.window = new Gtk.Window ();
        this.bind_property ("decorated", window, "decorated", BindingFlags.DEFAULT);

        this.application_settings = s;

        create_window_and_dialogs ();

        connect_signals ();
        
        inhibit_idle (true);
    }
    
    ~MainClientWindow ()
    {
        inhibit_idle (false);
    }

    void create_window_and_dialogs ()
    {
        mx_style = Mx.Style.get_default ();
        try {
            mx_style.load_from_file (GLib.Path.build_path (GLib.Path.DIR_SEPARATOR_S, BuildConfig.PACKAGE_DATADIR, BuildConfig.PACKAGE_NAME, "/style.css"));
        } catch (Error e) {
            warning ("Failed to load styles: %s", e.message);
        }

        window.focus_on_map = true;
        window.decorated = false;
        //window.skip_pager_hint = true;
        //window.skip_taskbar_hint = true;
        window.window_position = Gtk.WindowPosition.CENTER;
        window.gravity = Gdk.Gravity.CENTER;


        var screen = Gdk.Screen.get_default ();
        window.set_visual (screen.get_rgba_visual ());
        
        embed = new GtkClutter.Embed ();
        window.add (embed);
        embed.receives_default = (true);
        embed.can_default = (true);

        stage = embed.get_stage () as Clutter.Stage;
        stage.use_alpha = true;
        stage.color = {0, 0, 0, 0xFF};


        add_rtsp_client ();
        add_buttons ();


        this.settings_dialog = new SettingsDialog (this.stage, this.application_settings);
        this.peers_dialog = new PeersDialog (this.stage, this.application_settings);
        this.video_settings_dialog = new VideoSettingsDialog (this.stage, this.application_settings);
        this.oneclick_dialog = new OneClickStreamingDialog(this.stage, this.application_settings);

        this.secondaries = new Mx.BoxLayout ();
        this.secondaries.name ="secondaries";
        this.secondaries.orientation = Mx.Orientation.VERTICAL;
        this.layout.add (this.secondaries, Clutter.BinAlignment.END, Clutter.BinAlignment.START);
        this.application_settings.notify["rtsp-secondary-client-uris"].connect( () => {
            this.secondaries.foreach ( (c) => this.secondaries.remove(c) );
            foreach (var v in  this.application_settings.rtsp_secondary_client_uris)
                add_secondary (v);
        });
    }

    void add_rtsp_client ()
    {
        layout = new Clutter.BinLayout (Clutter.BinAlignment.CENTER, Clutter.BinAlignment.CENTER);
        var box = new Clutter.Box (layout);
        stage.add_actor (box);
        box.add_constraint (new Clutter.BindConstraint (stage, Clutter.BindCoordinate.SIZE, 0));


        video_texture = new Clutter.Texture ();
        layout.add (video_texture, Clutter.BinAlignment.CENTER, Clutter.BinAlignment.CENTER);
        video_texture.keep_aspect_ratio = true;

        var sink = new ClutterGst.VideoSink (video_texture);
        rtsp_client = new RTSPClient (sink);


        this.application_settings.bind_property ("rtsp-client-uri", rtsp_client, "uri", BindingFlags.BIDIRECTIONAL | BindingFlags.SYNC_CREATE);
        this.application_settings.bind_property ("is-muted", rtsp_client, "mute", BindingFlags.BIDIRECTIONAL | BindingFlags.SYNC_CREATE);
    }

    void add_buttons ()
    {
        var btns = new Mx.BoxLayout ();
        btns.name = "overlay-menu";

        var peebtn = new Mx.Button ();
        btns.add_actor (peebtn);
        peebtn.name = "peers-button";
        peebtn.tooltip_text = "View available peers.";
        peebtn.clicked.connect (() => {
            this.peers_dialog.show ();
        });

        var fulbtn = new Mx.Button ();
        btns.add_actor (fulbtn);
        fulbtn.name = "fullscreen-button";
        fulbtn.tooltip_text = "Toggle fullscreen mode.";
        fulbtn.clicked.connect (() => {
            if (!this.application_settings.is_fullscreen)
            {
              this.application_settings.is_fullscreen = true;
              fulbtn.style_pseudo_class_add ("enabled");
            }
            else
            {
              this.application_settings.is_fullscreen = false;
              fulbtn.style_pseudo_class_remove ("enabled");
            }
        });

        var vprebtn = new Mx.Button ();
        btns.add_actor (vprebtn);
        vprebtn.name = "video-preferences-button";
        vprebtn.tooltip_text = "Change video settings.";
        vprebtn.clicked.connect (() => {
            if (this.application_settings.use_advanced_video_preferences_dialog)
            {
              this.video_settings_dialog.show ();
            } else {
              this.oneclick_dialog.show ();
            }
        });

        var prebtn = new Mx.Button ();
        btns.add_actor (prebtn);
        prebtn.name = "preferences-button";
        prebtn.tooltip_text = "Change settings.";
        prebtn.clicked.connect (() => {
            this.settings_dialog.show ();
        });
        
        var cbtn = new Mx.Button ();
        btns.add_actor (cbtn);
        cbtn.name = "close-button";
        cbtn.tooltip_text = "Quit this application.";
        cbtn.clicked.connect (() => {
            rtsp_client.stop ();
            Gtk.main_quit ();
        });
        
        layout.add (btns, Clutter.BinAlignment.CENTER, Clutter.BinAlignment.END);
    }

    void connect_signals ()
    {
        window.delete_event.connect ( () => { 
            rtsp_client.stop ();
            Gtk.main_quit (); 
            return false; 
            });

        this.application_settings.notify.connect ( (sender, prop) => {
            switch (prop.name)
            {
                case "is-fullscreen":
                    if (this.application_settings.is_fullscreen)
                        this.window.fullscreen ();
                    else
                        this.window.unfullscreen ();
                    break;

                case "rtsp-client-enabled":
                    if (this.application_settings.rtsp_client_enabled == true && this.application_settings.rtsp_client_uri != null)
                    {
                        debug ("Enabling RTSP client.");
                        start ();
                    }
                    else
                    {
                        stop ();
                    }
                    break;
            }
        });
    }

    public new void show ()
    {
        var geometry = Gdk.Geometry ();
        geometry.min_width = 160;
        geometry.min_height = 90;
        geometry.min_aspect = 1.76; // aspects can not be the same, but should be around 16:9
        geometry.max_aspect = 1.78;
        window.set_geometry_hints (window, geometry, Gdk.WindowHints.MIN_SIZE | Gdk.WindowHints.ASPECT);
        float n = 0.75f;
        window.set_default_size ((int)(1280*n), (int)(720*n));
        
        window.show_all ();
    }

    public void run ()
    {
        Gtk.main ();
    }

    public void start ()
    {
        rtsp_client.start ();
    }
    
    public void stop ()
    {
        rtsp_client.stop ();
    }

    public void set_peers (HashTable<string,string> peers)
    {
        this.peers_dialog.set_peers (peers);
    }

    public void add_secondary (string uri)
    {
        debug ("Adding secondary: %s", uri);

        var w = new RTSPClientWidget (uri);
        w.tooltip_text = uri;
        w.closed.connect ( (s) => {
            this.secondaries.remove_actor (s);
            s.stop ();
        });

        this.secondaries.add (w);
    }
    
    void inhibit_idle (bool do_inhibit)
    {
        if (do_inhibit)
        {
            try
            {
              this.sessionmanager = Bus.get_proxy_sync (BusType.SESSION, 
                                                  "org.gnome.SessionManager", 
                                                  "/org/gnome/SessionManager");
              this.sessionmanager.Inhibit ("presence", 
                                        0,  //FIXME wtf!?
                                        "Awaiting or runing video presence.", 
                                        ISessionManager.InhibitFlags.IDLE,
                                        out this.sessionmanager_cookie);
            }
            catch (IOError error)
            {
              warning ("Could not inhibit idle: %s", error.message);
            }
       }
       else
       {
           this.sessionmanager.Uninhibit (this.sessionmanager_cookie);
       }
    }
}


class RTSPClientWidget : Mx.Frame
{
    public RTSPClient client;

    Clutter.BinLayout bin;
    Mx.BoxLayout buttons;
    Clutter.Texture tex;

    public signal void closed ();

    public RTSPClientWidget (string uri)
    {
        this.bin = new Clutter.BinLayout (Clutter.BinAlignment.CENTER, Clutter.BinAlignment.CENTER);
        this.child = new Clutter.Box (this.bin);

        this.tex = new Clutter.Texture ();
        this.tex.keep_aspect_ratio = true;
        this.tex.width = 200;

        var vs = new ClutterGst.VideoSink (tex);
        this.client = new RTSPClient.playing (vs, uri);
        this.bin.add (tex, Clutter.BinAlignment.CENTER, Clutter.BinAlignment.CENTER);

        create_buttons ();
    }

    void create_buttons ()
    {
        buttons = new Mx.BoxLayout ();
        buttons.name = "secondary-menu";
        
        var c = new Mx.Button ();
        c.name = "close-button";
        c.clicked.connect ( () => this.closed() );
        buttons.add (c);

        var m = new Mx.Button ();
        m.name = "mute-button";
        m.is_toggle = true;
        m.clicked.connect ( () => {
          if (m.toggled)
          {
            m.style_pseudo_class_add ("muted");
            m.tooltip_text = "Click to unmute";
            this.client.mute = true;
          }
          else
          {
            m.style_pseudo_class_remove ("muted");
            m.tooltip_text = "Click to mute";
            this.client.mute = false;
          }
        });
        buttons.add (m);

        this.bin.add (buttons, Clutter.BinAlignment.START, Clutter.BinAlignment.START);
    }

    public void stop ()
    {
        this.client.stop ();
    }
}


class SelectionExpander : Mx.Expander
{
  Mx.ScrollView scroll;
  Mx.BoxLayout box;

  public signal void selected (string entry);

  public struct Option
  {
    string name;
    string label;
  }

  public SelectionExpander (Option[] entries)
  {
    this.label = "(Default camera)";

    scroll = new Mx.ScrollView ();
    this.add_actor (scroll);

    box = new Mx.BoxLayout ();
    box.orientation = Mx.Orientation.VERTICAL;
    scroll.add_actor (box);

    for (int i = 0; i < entries.length; i++)
    {
      var entry = entries[i];
      var b = new Mx.Button.with_label (entry.label);
      b.clicked.connect ( (s) => {
          this.label = b.label;
          this.expanded = false;
          this.selected (entry.name);
      });
      box.add_actor (b);

      box.child_set_property (b, "x-fill", true);
      box.child_set_property (b, "expand", true);
    }
  }
}

[DBus (name = "org.gnome.SessionManager")]
interface ISessionManager : GLib.Object
{
  public abstract void Inhibit (string appname, uint xwinid, string reason, uint flags, out uint cookie) throws IOError;
  public abstract void Uninhibit (uint cookie) throws IOError;
  [Flags]
  public enum InhibitFlags
  {
    LOGOUT = 1,
    SWITCHING = 2,
    SUSPEND = 4,
    IDLE = 8
  }
}
